# Copyright 2006-2015 Mark Diekhans
"""
Robust, easy to use Unix process pipelines.
"""
import os
import signal
import shlex
import logging
import subprocess
import enum
from threading import RLock
from pipettor.devices import Dev
from pipettor.devices import DataReader
from pipettor.devices import _SiblingPipe
from pipettor.devices import File
from pipettor.exceptions import PipettorException
from pipettor.exceptions import ProcessException
from pipettor.exceptions import _warn_error_during_error_handling

# FIXME: C-c problems:
# http://code.activestate.com/recipes/496735-workaround-for-missed-sigint-in-multithreaded-prog/
# http://bugs.python.org/issue21822
# FIXME above handled by subprocess??

_defaultLogger = None
_defaultLogLevel = logging.DEBUG

def setDefaultLogger(logger):
    """Set the default pipettor logger used in logging command and errors.
    If None, there is no default logging.  The logger can be the name of
    a logger or the logger itself.  Standard value is None"""
    global _defaultLogger
    _defaultLogger = logging.getLogger(logger) if isinstance(logger, str) else logger

def getDefaultLogger():
    """return the current value of the pipettor default logger"""
    return _defaultLogger

def setDefaultLogLevel(level):
    """Set the default pipettor log level to use in logging command and errors.
    Standard value is logging.DEBUG"""
    global _defaultLogLevel
    _defaultLogLevel = level

def getDefaultLogLevel():
    """Get the default pipettor log level to use in logging command and errors."""
    return _defaultLogLevel

def setDefaultLogging(logger, level):
    """Set both default logger and level. Either can be None to leave as default"""
    if logger is not None:
        setDefaultLogger(logger)
    if level is not None:
        setDefaultLogLevel(level)

def _getLoggerToUse(logger):
    """if logger is None, get default, otherwise if it's a string, look it up,
    otherwise it's the logger object."""
    if logger is None:
        return _defaultLogger
    elif isinstance(logger, str):
        return logging.getLogger(logger)
    else:
        return logger

def _getLogLevelToUse(logLevel):
    "get log level to use, either what is specified or default"
    return logLevel if logLevel is not None else getDefaultLogLevel()

class State(enum.IntEnum):
    """Current state of a process"""
    PREINIT = 0
    STARTUP = 1
    RUNNING = 2
    FINISHED = 4

class Process(object):
    """A process, represented as a node a pipeline, connected by Dev objects.

    Process arguments can be can be any object, with str() being called on
    the object before exec.

    If the stdin/out/err arguments can have the following values:
       - None - stdio file descriptior is inherited.
       - str - opened as a file
       - int -  file number
       - file-like object - fileno() is dupped
       - a Dev derived object

    If stderr is an instance of DataReader, then stderr is included in
    ProcessException on process error.  If the class DataReader is passed
    in as stderr, a DataReader object is created.

    start() must be called to run process
    """

    def __init__(self, cmd, stdin=None, stdout=None, stderr=None):
        self.lock = RLock()
        self.cmd = tuple(cmd)
        # stdio and argument Dev association
        self.stdin = self._stdio_assoc(stdin, "r")
        self.stdout = self._stdio_assoc(stdout, "w")
        if stderr == DataReader:
            stderr = DataReader(errors='backslashreplace')
        self.stderr = self._stdio_assoc(stderr, "w")
        self.popen = None
        self.pid = None
        self.pgid = None
        self.returncode = None  # exit code, or -signal
        # FIXME: should this just be exception for the users??
        self.procExcept = None  # exception because of failed process
        self.state = State.PREINIT
        self.forced = False    # force termination during error cleanup

    def __str__(self):
        "get simple description of process"
        return " ".join([shlex.quote(str(arg)) for arg in self.cmd])

    def _stdio_assoc(self, spec, mode):
        """pre-fork check a stdio spec validity and associate Dev or file
        number.  mode is mode in child"""
        if (spec is None) or isinstance(spec, int):
            return spec  # passed unchanged
        elif isinstance(spec, Dev):
            spec._bind_to_process(self, mode)
            return spec  # passed unchanged
        elif callable(getattr(spec, "fileno", None)):
            return spec.fileno()  # is file-like
        elif isinstance(spec, str):
            return File(spec, mode)
        else:
            raise PipettorException("invalid stdio specification object type: {} {}".format(type(spec), spec))

    def _get_child_stdio(self, spec, stdfd):
        """get fd to pass to child as one of the stdio handles."""
        if spec is None:
            return stdfd
        elif isinstance(spec, int):
            return spec
        elif isinstance(spec, Dev):
            return spec.read_fd if stdfd == 0 else spec.write_fd
        else:
            # this should have been detected earlier
            raise PipettorException("_get_child_stdio logic error: {} {}".format(type(spec), stdfd))

    def _start_process(self, pgid):
        """Do work of starting the process.  If pgid is None, this process
        becomes group leader, otherwise this process is added to group pgid."""
        self.state = State.STARTUP    # do first to prevent restarts on error

        # standard dance to get process group set
        # preexec_fn will go away: https://bugs.python.org/issue38435
        if pgid is None:
            preexecFn = lambda: os.setpgid(os.getpid(), os.getpid())  # noqa
        else:
            preexecFn = lambda: os.setpgid(os.getpid(), pgid)  # noqa

        try:
            self.popen = subprocess.Popen(self.cmd,
                                          stdin=self._get_child_stdio(self.stdin, 0),
                                          stdout=self._get_child_stdio(self.stdout, 1),
                                          stderr=self._get_child_stdio(self.stderr, 2),
                                          preexec_fn=preexecFn)
        except Exception as ex:
            raise ProcessException(str(self)) from ex
        self.pid = self.popen.pid
        self.pgid = self.pid if pgid is None else pgid
        self.state = State.RUNNING

    def _start(self, pgid):
        """Start the process,  If pgid is None, this process
        becomes group leader."""
        try:
            self._start_process(pgid)
        except BaseException as ex:
            self.procExcept = ex
        if self.procExcept is not None:
            raise self.procExcept

    @property
    def running(self):
        "determined if this process has been running"
        return self.state is State.RUNNING

    @property
    def finished(self):
        "determined if been detected as finished (waited on)"
        return self.state is State.FINISHED

    def _parent_stdio_exit_close(self):
        "close devices on edit"
        # MUST do before reading stderr in _handle_error_exit
        for std in (self.stdin, self.stdout, self.stderr):
            if isinstance(std, Dev):
                std.close()

    def _handle_error_exit(self):
        # get saved stderr, if possible
        stderr = None
        if isinstance(self.stderr, DataReader):
            stderr = self.stderr.data
        # don't save exception if we force it to be killed
        if not self.forced:
            self.procExcept = ProcessException(str(self), self.returncode, stderr)

    def _handle_exit(self, waitStat):
        """Handle process exiting, saving status  """
        self.state = State.FINISHED
        assert os.WIFEXITED(waitStat) or os.WIFSIGNALED(waitStat)
        self.returncode = os.WEXITSTATUS(waitStat) if os.WIFEXITED(waitStat) else -os.WTERMSIG(waitStat)
        # must tell subprocess.Popen about this
        self.popen.returncode = self.returncode
        self._parent_stdio_exit_close()  # MUST DO BEFORE _handle_error_exit
        if not ((self.returncode == 0) or (self.returncode == -signal.SIGPIPE)):
            self._handle_error_exit()

    def _waitpid(self, flag=0):
        "Do waitpid and handle exit if finished, return True if finished"
        if self.pid is None:
            raise PipettorException("process has not been started")
        w = os.waitpid(self.pid, flag)
        if w[0] != 0:
            self._handle_exit(w[1])
        return (w[0] != 0)

    def poll(self):
        """Check if the process has completed.  Return True if it
        has, False if it hasn't."""
        with self.lock:
            if self.state is State.FINISHED:
                return True
            return self._waitpid(os.WNOHANG)

    def _force_finish(self):
        """Force termination of process.  The forced flag is set, as an
        indication that this was not a primary failure in the pipeline.
        """
        with self.lock:
            # check if finished before killing
            if (self.state is State.RUNNING) and (not self.poll()):
                self.forced = True
                os.kill(self.pid, signal.SIGKILL)
                self._waitpid()

    def failed(self):
        "check if process failed, call after poll() or wait()"
        return self.procExcept is not None


class Pipeline(object):
    """
    A process pipeline.  Once constructed, the pipeline
    is started with start(), poll(), or wait() functions.

    The cmds argument is either a list of arguments for a single process, or a
    list of such lists for a pipeline. If the stdin/out/err arguments are
    none, the open files are are inherited.  Otherwise they can be string file
    names, file-like objects, file number, or Dev object.  Stdin is input to
    the first process, stdout is output to the last process and stderr is
    attached to all processed. DataReader and DataWriter objects can be
    specified for stdin/out/err asynchronously I/O with the pipeline without
    the danger of deadlock.

    If stderr is the class DataReader, a new instance is created for each
    process in the pipeline. The contents of stderr will include an
    exception if an occurs in that process.  If an instance of DataReader
    is provided, the contents of stderr from all process will be included in
    the exception.

    Command arguments will be converted to strings.

    The logger argument can be the name of a logger or a logger object.  If
    none, default is user.
    """
    def __init__(self, cmds, stdin=None, stdout=None, stderr=DataReader,
                 logger=None, logLevel=None):
        self.lock = RLock()
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.procs = []
        self.devs = set()
        self.pgid = None       # process group leader
        self.bypid = dict()    # indexed by pid
        self.state = State.PREINIT
        self.logger = _getLoggerToUse(logger)
        self.logLevel = _getLogLevelToUse(logLevel)

        if isinstance(cmds[0], str):
            cmds = [cmds]  # one-process pipeline
        cmds = self._stringify(cmds)
        try:
            self._setup_processes(cmds)
        except BaseException:
            self._error_cleanup()
            raise

    @staticmethod
    def _stringify(cmds):
        ncmds = []
        for cmd in cmds:
            ncmds.append([str(a) for a in cmd])
        return ncmds

    @property
    def running(self):
        "determined if this process has been running"
        return self.state is State.RUNNING

    @property
    def finished(self):
        "determined if been detected as finished (waited on)"
        return self.state is State.FINISHED

    def _log(self, level, message, ex=None):
        """If logging is available and enabled, log message and optional
        exception"""
        if (self.logger is not None) and (self.logger.isEnabledFor(level)):
            kwargs = {}
            if ex is not None:
                kwargs["exc_info"] = ex
            self.logger.log(level, "{}: {}".format(message, str(self)), **kwargs)

    def _setup_processes(self, cmds):
        prevPipe = None
        lastCmdIdx = len(cmds) - 1
        for i in range(len(cmds)):
            prevPipe = self._add_process(cmds[i], prevPipe, (i == lastCmdIdx), self.stdin, self.stdout, self.stderr)

    def _add_process(self, cmd, prevPipe, isLastCmd, stdinFirst, stdoutLast, stderr):
        """add one process to the pipeline, return the output pipe if not the last process"""
        if prevPipe is None:
            stdin = stdinFirst  # first process in pipeline
        else:
            stdin = prevPipe
        if isLastCmd:
            outPipe = None
            stdout = stdoutLast  # last process in pipeline
        else:
            outPipe = stdout = _SiblingPipe()
        try:
            self._create_process(cmd, stdin, stdout, stderr)
        except BaseException:
            if outPipe is not None:
                outPipe.close()
            raise
        return outPipe

    def _create_process(self, cmd, stdin, stdout, stderr):
        """create process and track Dev objects"""
        proc = Process(cmd, stdin, stdout, stderr)
        self.procs.append(proc)
        # Proc maybe have wrapped a Dev
        for std in (proc.stdin, proc.stdout, proc.stderr):
            if isinstance(std, Dev):
                self.devs.add(std)

    def __str__(self):
        """get a string describing the pipe"""
        desc = str(self.procs[0])
        if self.stdin not in (None, 0):
            desc += " <" + str(self.stdin)
        if len(self.procs) > 1:
            desc += " | " + " | ".join([str(proc) for proc in self.procs[1:]])
        if self.stdout not in (None, 1):
            desc += " >" + str(self.stdout)
        if self.stderr == DataReader:
            desc += " 2>[DataReader]"  # instance made in Process
        elif self.stderr not in (None, 2):
            desc += " 2>" + str(self.stderr)
        return desc

    def _start_process(self, proc):
        proc._start(self.pgid)
        self.bypid[proc.pid] = proc
        self.pgid = proc.pgid

    def _start_processes(self):
        for proc in self.procs:
            self._start_process(proc)

    def _post_start_parent(self):
        for d in self.devs:
            d._post_start_parent()

    def _finish(self):
        "finish up when no errors have occurred"
        assert not self.failed()
        self.state = State.FINISHED
        for d in self.devs:
            d.close()
        self._log(self.logLevel, "success")

    def _log_failure(self, ex):
        self._log(logging.ERROR, "failure", ex)

    def _error_cleanup_dev(self, dev):
        try:
            dev.close()
        except Exception as ex:
            _warn_error_during_error_handling("error during device cleanup on error", ex)

    def _error_cleanup_process(self, proc):
        try:
            if not proc.finished:
                proc._force_finish()
        except Exception as ex:
            _warn_error_during_error_handling("error during process cleanup on error", ex)

    def _error_cleanup(self):
        """forced cleanup of child processed after failure"""
        self.state = State.FINISHED
        for d in self.devs:
            self._error_cleanup_dev(d)
        for p in self.procs:
            self._error_cleanup_process(p)

    def _start_guts(self):
        self._log(self.logLevel, "start")
        self.state = State.RUNNING
        # clean up devices and process if there is a failure
        try:
            self._start_processes()
            self._post_start_parent()
        except Exception as ex:
            self._log_failure(ex)
            self._error_cleanup()
            raise

    def start(self):
        """start processes"""
        if self.state >= State.STARTUP:
            raise PipettorException("Pipeline is already been started")
        with self.lock:
            self._start_guts()

    def _raise_if_failed(self):
        """raise exception if any process has one, otherwise do nothing"""
        try:
            for p in self.procs:
                if p.procExcept is not None:
                    raise p.procExcept
        except Exception as ex:
            self._log_failure(ex)
            raise ex

    def _poll_guts(self):
        for p in self.procs:
            if not p.poll():
                return False
        self._finish()
        return True

    def poll(self):
        """Check if all of the processes have completed.  Return True if it
        has, False if it hasn't."""
        with self.lock:
            if self.state is State.PREINIT:
                self.start()
            try:
                return self._poll_guts()
            except BaseException:
                self._error_cleanup()
                raise

    def _wait_on_one(self, proc):
        "wait on the next process in group to complete"
        w = os.waitpid(proc.pid, 0)
        self.bypid[w[0]]._handle_exit(w[1])

    def _wait_guts(self):
        if self.state < State.RUNNING:
            self.start()
        try:
            for p in self.procs:
                if not p.finished:
                    self._wait_on_one(p)
        except BaseException as ex:
            self._log_failure(ex)
            self._error_cleanup()
            raise ex
        self._raise_if_failed()

    def wait(self):
        """Wait for all of the process to complete. Generate an exception if
        any exits non-zero or signals. Starts process if not already
        running."""
        with self.lock:
            self._wait_guts()
            self._finish()

    def _shutdown(self):
        "guts of shutdown"
        self.kill(sig=signal.SIGKILL)
        try:
            self.wait()
        except PipettorException:
            pass  # ignore errors we report

    def shutdown(self):
        """Close down the pipeline prematurely. If the pipeline is running,
        it's killed.  This does not report errors from child process and
        differs from wait in the fact that it doesn't start the pipeline if it
        has not been started, just frees up open pipes. Primary intended
        for error recovery"""
        with self.lock:
            if self.logger is not None:
                self.logger.log(self.logLevel, "Shutting down pipeline: {}".format(str(self)))
            if self.state is State.RUNNING:
                self._shutdown()
            elif self.state is not State.FINISHED:
                self._finish()  # just clean up pipes

    def failed(self):
        "check if any process failed, call after poll() or wait()"
        with self.lock:
            for p in self.procs:
                if p.failed():
                    return True
            return False

    def kill(self, sig=signal.SIGTERM):
        "send a signal to all of the processes in the pipeline"
        for p in self.procs:
            os.kill(p.pid, sig)


class Popen(Pipeline):
    """File-like object of processes to read from or write to a Pipeline.

    The cmds argument is either a list of arguments for a single process,
    or a list of such lists for a pipeline.  Mode is 'r' for a pipeline
    who's output will be read, or 'w' for a pipeline to that is to have
    data written to it.  If stdin or stdout is specified, and is a string,
    it is a file to open as other file at the other end of the pipeline.
    If it's not a string, it is assumed to be a file object to use for
    input or output.  For a read pipe, only stdin can be specified, for a
    write pipe, only stdout can be used.

    read pipeline ('r'):
      stdin --> cmd[0] --> ... --> cmd[n] --> Popen

    write pipeline ('w')
      Popen --> cmd[0] --> ... --> cmd[n] --> stdout

    Command arguments will be converted to strings.

    The logger argument can be the name of a logger or a logger object.  If
    none, default is user.

    Specifying binary access results in data of type bytes, otherwise str type
    is returned.  The buffering, encoding, and errors arguments are as used in
    the open() function.
    """

    def __init__(self, cmds, mode='r', stdin=None, stdout=None, logger=None, logLevel=None,
                 buffering=-1, encoding=None, errors=None):
        self.mode = mode
        self._parent_fh = None
        self._child_fd = None
        if mode.find('a') >= 0:
            raise PipettorException("can not specify append mode")
        if mode.find('r') >= 0:
            if stdout is not None:
                raise PipettorException("can not specify stdout with read mode")
        else:
            if stdin is not None:
                raise PipettorException("can not specify stdin with write mode")

        pipe_read_fd, pipe_write_fd = os.pipe()
        if mode.find('r') >= 0:
            firstIn = stdin
            lastOut = pipe_write_fd
            self._child_fd = pipe_write_fd
            self._parent_fh = open(pipe_read_fd, mode, buffering=buffering, encoding=encoding, errors=errors)
        else:
            firstIn = pipe_read_fd
            lastOut = stdout
            self._child_fd = pipe_read_fd
            self._parent_fh = open(pipe_write_fd, mode, buffering=buffering, encoding=encoding, errors=errors)
        super(Popen, self).__init__(cmds, stdin=firstIn, stdout=lastOut, logger=logger, logLevel=logLevel)
        self.start()
        os.close(self._child_fd)
        self._child_fd = None

    def _close(self):
        if self._parent_fh is not None:
            self._parent_fh.close()
            self._parent_fh = None
        if self._child_fd is not None:
            os.close(self._child_fd)
            self._child_fd = None

    def __enter__(self):
        "support for with statement"
        return self

    def __exit__(self, type, value, traceback):
        "support for with statement"
        self.close()

    def __iter__(self):
        "iter over contents of file"
        return self._parent_fh.__iter__()

    def next(self):
        return self._parent_fh.next()

    def flush(self):
        "Flush the internal I/O buffer."
        self._parent_fh.flush()

    def fileno(self):
        "get the integer OS-dependent file handle"
        return self._parent_fh.fileno()

    def write(self, str):
        "Write string str to file."
        self._parent_fh.write(str)

    def read(self, size=-1):
        return self._parent_fh.read(size)

    def readline(self, size=-1):
        return self._parent_fh.readline(size)

    def readlines(self, size=-1):
        return self._parent_fh.readlines(size)

    def wait(self):
        """wait to for processes to complete, generate an exception if one
        exits no-zero"""
        with self.lock:
            self._close()
            super(Popen, self).wait()

    def poll(self):
        "poll is not allowed for Pipeline objects"
        # FIXME: don't know what to do about our open pipe keeping process from
        # exiting so we can get a status, so disallow it. Not sure how to
        # address this.  Can probably address this with select on pipe.
        raise PipettorException("Pipeline.poll() is not supported")

    def close(self):
        "wait for process to complete, with an error if it exited non-zero"
        with self.lock:
            self._close()
            if self.state is not State.FINISHED:
                self.wait()



# Copyright 2006-2016 Mark Diekhans
import sys
import traceback
import signal
from warnings import warn


def _signal_num_to_name(num):
    "get name for a signal number"
    # find name in signal namespace
    for key in vars(signal):
        if (getattr(signal, key) == num) and key.startswith("SIG") and (key.find("_") < 0):
            return key
    return "signal" + str(num)


class PipettorException(Exception):
    """Base class for Pipettor exceptions."""
    pass


class ProcessException(PipettorException):
    """Exception associated with running a process.  A None returncode indicates
    a exec failure."""
    def __init__(self, procDesc, returncode=None, stderr=None):
        self.procDesc = procDesc
        self.returncode = returncode
        self.stderr = stderr
        if returncode is None:
            msg = "exec failed"
        elif (returncode < 0):
            msg = "process signaled: " + _signal_num_to_name(-returncode)
        else:
            msg = "process exited " + str(returncode)
        if procDesc is not None:
            msg += ": " + procDesc
        if (stderr is not None) and (len(stderr) != 0):
            msg += ":\n" + stderr
        super(ProcessException, self).__init__(msg)

    def __reduce__(self):
        # __reduce__ is used, otherwise we had problems with the msg having
        # duplicated string
        return (ProcessException, (self.procDesc, self.returncode, self.stderr))


class ErrorDuringErrorHandlingWarning(Warning):
    """An error occurred while handing another error"""
    pass


def _warn_error_during_error_handling(msg, exception):
    "called to issue warning on error during error handling"
    exi = sys.exc_info()
    stack = "" if exi is None else "".join(traceback.format_list(traceback.extract_tb(exi[2]))) + "\n"
    warn(msg + " " + str(exception) + "\n" + stack, ErrorDuringErrorHandlingWarning)



"""
Robust, easy to use Unix process pipelines.
"""
import shlex
from pipettor.exceptions import PipettorException, ProcessException
from pipettor.devices import DataReader, DataWriter, File
from pipettor.processes import Pipeline, Popen, setDefaultLogger, getDefaultLogger, setDefaultLogLevel, getDefaultLogLevel, setDefaultLogging

__version__ = "0.7.0"


def run(cmds, stdin=None, stdout=None, stderr=DataReader, logger=None, logLevel=None):
    """Construct and run an process pipeline. If any of the processes fail,
    a ProcessException is throw.

    `cmds` is either a list of arguments for a single process, or a list of
    such lists for a pipeline. If the `stdin`, `stdout`, or `stderr` arguments
    are none, the open files are are inherited.  Otherwise they can be string
    file names, file-like objects, file number, or :class:`pipettor.Dev`
    object.  `stdin` is input to the first process, `stdout` is output to the
    last process and `stderr` is attached to all processed.
    :class:`pipettor.DataReader` and :class:`pipettor.DataWriter` objects can
    be specified for `stdin`, `stdout`, or `stderr` asynchronously I/O with
    the pipeline without the danger of deadlock.

    If stderr is the class DataReader, a new instance is created for each
    process in the pipeline. The contents of stderr will include an exception
    if an occurs in that process.  If an instance of
    :class:`pipettor.DataReader` is provided, the contents of stderr from all
    process will be included in the exception.
    """
    Pipeline(cmds, stdin=stdin, stdout=stdout, stderr=stderr, logger=logger, logLevel=logLevel).wait()


def runout(cmds, stdin=None, stderr=DataReader, logger=None, logLevel=None,
           buffering=-1, encoding=None, errors=None):
    """Construct and run an process pipeline, returning the output. If any of the
    processes fail, a ProcessException is throw.

    See the :func:`pipettor.run` function for more details.  Use
    `str.splitlines()` to split result into lines.

    The logger argument can be the name of a logger or a logger object.  If
    none, default is user.

    Specifying binary access results in data of type bytes, otherwise str type
    is return.  The buffering, encoding, and errors arguments are as used in
    the open() function.
    """
    dr = DataReader(buffering=buffering, encoding=encoding, errors=errors)
    Pipeline(cmds, stdin=stdin, stdout=dr, stderr=stderr, logger=logger, logLevel=logLevel).wait()
    return dr.data


def _lexcmds(cmds):
    """spit pipeline specification into arguments"""
    if isinstance(cmds, str):
        return shlex.split(cmds)
    else:
        return [shlex.split(cmd) if isinstance(cmd, str) else cmd for cmd in cmds]


def runlex(cmds, stdin=None, stdout=None, stderr=DataReader, logger=None, logLevel=None):
    """Call :func:`pipettor.run`, first splitting commands specified as strings
    are split into arguments using `shlex.split`.

    If `cmds` is a string, it is split into arguments and run as as a single
    process.  If `cmds` is a list, a multi-process pipeline is created.
    Elements that are strings are split into arguments to form commands.
    Elements that are lists are treated as commands without splitting.
    """
    run(_lexcmds(cmds), stdin=stdin, stdout=stdout, stderr=stderr, logger=logger, logLevel=None)


def runlexout(cmds, stdin=None, stderr=DataReader, logger=None, logLevel=None,
              buffering=-1, encoding=None, errors=None):
    """Call :func:`pipettor.runout`, first splitting commands specified
    as strings are split into arguments using `shlex.split`.

    If `cmds` is a string, it is split into arguments and run as as a single
    process.  If `cmds` is a list, a multi-process pipeline is created.
    Elements that are strings are split into arguments to form commands.
    Elements that are lists are treated as commands without splitting.

    The logger argument can be the name of a logger or a logger object.  If
    none, default is user.

    Specifying binary access results in data of type bytes, otherwise str type
    is returned.  The buffering, encoding, and errors arguments are as used in
    the open() function.
    """
    return runout(_lexcmds(cmds), stdin=stdin, stderr=stderr, logger=logger, logLevel=logLevel,
                  buffering=buffering, encoding=encoding, errors=errors)


# n.b. all of the library API functions and classes need to be explicitly
# included in the docs/library.rst files
__all__ = (PipettorException.__name__, ProcessException.__name__,
           DataReader.__name__, DataWriter.__name__,
           File.__name__, Pipeline.__name__, Popen.__name__,
           setDefaultLogger.__name__, getDefaultLogger.__name__,
           setDefaultLogLevel.__name__, getDefaultLogLevel.__name__, setDefaultLogging.__name__,
           run.__name__, runout.__name__, runlex.__name__, runlexout.__name__)



# -*- coding: utf-8 -*-
# Copyright 2006-2015 Mark Diekhans
import unittest
import sys
import os
import re
import signal
if __name__ == '__main__':
    sys.path.insert(0, os.path.normpath(os.path.dirname(sys.argv[0])) + "/../lib")
    from testCaseBase import TestCaseBase, LoggerForTests
else:
    from .testCaseBase import TestCaseBase, LoggerForTests
from pipettor import Pipeline, Popen, ProcessException, PipettorException, DataReader, DataWriter, File, run, runout, runlex, runlexout


def sigquit_handler(signum, frame):
    " prevent MacOS  crash reporter"
    sys.exit(os.EX_SOFTWARE)


signal.signal(signal.SIGQUIT, sigquit_handler)

# this keeps OS/X crash reporter from popping up on unittest error
signal.signal(signal.SIGQUIT,
              lambda signum, frame: sys.exit(os.EX_SOFTWARE))
signal.signal(signal.SIGABRT,
              lambda signum, frame: sys.exit(os.EX_SOFTWARE))


class PipettorTestBase(TestCaseBase):
    "provide common functions used in test classes"
    def __init__(self, methodName):
        super(PipettorTestBase, self).__init__(methodName)

    def orphanChecks(self, nopen):
        "check for orphaned child process, open files or threads"
        self.assertNoChildProcs()
        self.assertNumOpenFilesSame(nopen)
        self.assertSingleThread()

    def commonChecks(self, nopen, pipeline, expectStr, isRe=False):
        """check that files, threads, and processes have not leaked. Check str(pipeline)
        against expectStr, which can be a string, or an regular expression if
        isRe==True, or None to not check."""
        s = str(pipeline)
        if expectStr is not None:
            if isRe:
                if not re.search(expectStr, s):
                    self.fail("'" + s + "' doesn't match RE '" + expectStr + "'")
            else:
                self.assertEqual(s, expectStr)
        self.orphanChecks(nopen)

    def checkProgWithError(self, procExcept, progArgs=None):
        expectReTmpl = "^process exited 1: .+/progWithError{}{}:\nTHIS GOES TO STDERR{}{}.*$"
        if progArgs is not None:
            expectRe = expectReTmpl.format(" ", progArgs, ": ", progArgs)
        else:
            expectRe = expectReTmpl.format("", "", "", "")
        if not re.match(expectRe, str(procExcept), re.MULTILINE):
            self.fail("'{}' does not match '{}'".format(str(procExcept), str(expectRe)))


class PipelineTests(PipettorTestBase):
    def __init__(self, methodName):
        super(PipelineTests, self).__init__(methodName)

    def testTrivial(self):
        nopen = self.numOpenFiles()
        pl = Pipeline(("true",))
        pl.wait()
        self.commonChecks(nopen, pl, "true 2>[DataReader]")

    def testTrivialPoll(self):
        nopen = self.numOpenFiles()
        pl = Pipeline(("sleep", "1"))
        while not pl.poll():
            pass
        pl.wait()
        self.commonChecks(nopen, pl, "sleep 1 2>[DataReader]")

    def testTrivialFailPoll(self):
        nopen = self.numOpenFiles()
        pl = Pipeline([("sleep", "1"), ("false",)])
        with self.assertRaisesRegex(ProcessException, "^process exited 1: sleep 1 | false$"):
            pl.wait()
        self.commonChecks(nopen, pl, "sleep 1 | false 2>[DataReader]")

    def testTrivialStatus(self):
        nopen = self.numOpenFiles()
        pl = Pipeline(("true",))
        pl.start()
        self.assertTrue(pl.running)
        self.assertFalse(pl.finished)
        pl.wait()
        self.assertFalse(pl.running)
        self.assertTrue(pl.finished)
        self.commonChecks(nopen, pl, "true 2>[DataReader]")

    def testSimplePipe(self):
        nopen = self.numOpenFiles()
        log = LoggerForTests()
        pl = Pipeline([("true",), ("true",)], logger=log.logger)
        pl.wait()
        self.commonChecks(nopen, pl, "true | true 2>[DataReader]")
        self.assertEqual(log.data, """start: true | true 2>[DataReader]\n"""
                         """success: true | true 2>[DataReader]\n""")

    def testSimplePipeFail(self):
        nopen = self.numOpenFiles()
        log = LoggerForTests()
        pl = Pipeline([("false",), ("true",)], logger=log.logger)
        with self.assertRaisesRegex(ProcessException, "^process exited 1: false$"):
            pl.wait()
        self.commonChecks(nopen, pl, "false | true 2>[DataReader]")
        self.assertRegex(log.data,
                         re.compile("""^start: false | true 2>[DataReader]\n"""
                                    """failure: false | true 2>[DataReader]: process exited 1: false\n.*""",
                                    re.MULTILINE))

    def testPipeFailStderr(self):
        nopen = self.numOpenFiles()
        # should report first failure
        pl = Pipeline([("true",), (os.path.join(self.getTestDir(), "progWithError"),), ("false",)], stderr=DataReader)
        with self.assertRaises(ProcessException) as cm:
            pl.wait()
        self.checkProgWithError(cm.exception)
        self.orphanChecks(nopen)

    def testPipeFail3Stderr(self):
        # all 3 process fail
        nopen = self.numOpenFiles()
        # should report first failure
        pl = Pipeline([(os.path.join(self.getTestDir(), "progWithError"), "process0"),
                       (os.path.join(self.getTestDir(), "progWithError"), "process1"),
                       (os.path.join(self.getTestDir(), "progWithError"), "process2")],
                      stderr=DataReader)
        with self.assertRaises(ProcessException) as cm:
            pl.wait()
        # should be first process
        self.checkProgWithError(cm.exception, "process0")
        # check process
        for i in range(3):
            self.checkProgWithError(pl.procs[i].procExcept, "process{}".format(i))
        self.orphanChecks(nopen)

    def testExecFail(self):
        # invalid executable
        nopen = self.numOpenFiles()
        dw = DataWriter("one\ntwo\nthree\n")
        pl = Pipeline(("procDoesNotExist", "-r"), stdin=dw)
        with self.assertRaises(ProcessException) as cm:
            pl.wait()
        self.assertRegex(str(cm.exception),
                         "exec failed: procDoesNotExist -r.*")
        self.assertIsNot(cm.exception.__cause__, None)
        self.assertRegex(str(cm.exception.__cause__),
                         "\\[Errno 2\\] No such file or directory: 'procDoesNotExist'.*")
        self.commonChecks(nopen, pl, "procDoesNotExist -r <[DataWriter] 2>[DataReader]")

    def testSignaled(self):
        # process signals
        nopen = self.numOpenFiles()
        pl = Pipeline(("sh", "-c", "kill -11 $$"))
        with self.assertRaises(ProcessException) as cm:
            pl.wait()
        expect = "process signaled: SIGSEGV: sh -c 'kill -11 $$'"
        msg = str(cm.exception)
        if not msg.startswith(expect):
            self.fail("'" + msg + "' does not start with '"
                      + expect + "', cause: " + str(getattr(cm.exception, "cause", None)))
        self.commonChecks(nopen, pl, "sh -c 'kill -11 $$' 2>[DataReader]")

    def testStdinMem(self):
        # write from memory to stdin
        nopen = self.numOpenFiles()
        outf = self.getOutputFile(".out")
        dw = DataWriter("one\ntwo\nthree\n")
        pl = Pipeline(("sort", "-r"), stdin=dw, stdout=outf)
        pl.wait()
        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "^sort -r <\\[DataWriter\\] >.+/output/test_pipettor\\.PipelineTests\\.testStdinMem\\.out 2>\\[DataReader\\]$", isRe=True)

    def testStdoutMem(self):
        # read from stdout into memory
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        dr = DataReader()
        pl = Pipeline(("sort", "-r"), stdin=inf, stdout=dr)
        pl.wait()
        self.assertEqual(dr.data, "two\nthree\nsix\none\nfour\nfive\n")
        self.commonChecks(nopen, pl, "^sort -r <.+/input/simple1\\.txt >\\[DataReader\\]", isRe=True)

    def testStdinStdoutMem(self):
        # write and read from memory
        nopen = self.numOpenFiles()
        dw = DataWriter("one\ntwo\nthree\n")
        dr = DataReader()
        pl = Pipeline([("cat", "-u"), ("cat", "-u")], stdin=dw, stdout=dr)
        pl.wait()
        self.assertEqual(dr.data, "one\ntwo\nthree\n")
        self.commonChecks(nopen, pl, "^cat -u <\\[DataWriter\\] \\| cat -u >\\[DataReader\\] 2>\\[DataReader\\]$", isRe=True)

    def testFileMode(self):
        with self.assertRaisesRegex(PipettorException, "^invalid or unsupported mode 'q' opening /dev/null"):
            File("/dev/null", "q")

    def testCollectStdoutErr(self):
        # independent collection of stdout and stderr
        nopen = self.numOpenFiles()
        stdoutRd = DataReader()
        stderrRd = DataReader()
        pl = Pipeline(("sh", "-c", "echo this goes to stdout; echo this goes to stderr >&2"),
                      stdout=stdoutRd, stderr=stderrRd)
        pl.wait()
        self.assertEqual(stdoutRd.data, "this goes to stdout\n")
        self.assertEqual(stderrRd.data, "this goes to stderr\n")
        self.commonChecks(nopen, pl, "sh -c 'echo this goes to stdout; echo this goes to stderr >&2' >[DataReader] 2>[DataReader]")

    def testStdinMemBinary(self):
        # binary write from memory to stdin
        nopen = self.numOpenFiles()
        outf = self.getOutputFile(".out")
        fh = open(self.getInputFile("file.binary"), "rb")
        dw = DataWriter(fh.read())
        fh.close()
        pl = Pipeline(("cat",), stdin=dw, stdout=outf)
        pl.wait()
        self.diffBinaryExpected(".out", expectedBasename="file.binary")
        self.commonChecks(nopen, pl, "^cat <\\[DataWriter] >.*/output/test_pipettor.PipelineTests.testStdinMemBinary.out 2>\\[DataReader\\]$", isRe=True)

    def testStdoutMemBinary(self):
        # binary read from stdout into memory
        nopen = self.numOpenFiles()
        inf = self.getInputFile("file.binary")
        dr = DataReader(binary=True)
        pl = Pipeline(("cat",), stdin=inf, stdout=dr)
        pl.wait()
        fh = open(self.getOutputFile(".out"), "wb")
        fh.write(dr.data)
        fh.close()
        self.diffBinaryExpected(".out", expectedBasename="file.binary")
        self.commonChecks(nopen, pl, "^cat <.*/input/file.binary >\\[DataReader] 2>\\[DataReader\\]$", isRe=True)

    def testWriteFile(self):
        # test write to File object
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        outf = self.getOutputFile(".out")
        # double cat actually found a bug
        pl = Pipeline([("cat",), ("cat",)], stdin=inf, stdout=File(outf, "w"))
        pl.wait()
        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "cat <.*/input/simple1.txt \\| cat >.*/output/test_pipettor.PipelineTests.testWriteFile.out 2>\\[DataReader\\]$", isRe=True)

    def testReadFile(self):
        # test read and write to File object
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        outf = self.getOutputFile(".out")
        pl = Pipeline([("cat",), ("cat",)], stdin=File(inf), stdout=File(outf, "w"))
        pl.wait()
        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "cat <.*/input/simple1.txt \\| cat >.*/output/test_pipettor.PipelineTests.testReadFile.out 2>\\[DataReader\\]$", isRe=True)

    def testAppendFile(self):
        # test append to File object
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        outf = self.getOutputFile(".out")
        # double cat actually found a bug
        pl = Pipeline([("cat",), ("cat",)], stdin=inf, stdout=File(outf, "w"), stderr=None)
        pl.wait()
        pl = Pipeline([("cat",), ("cat",)], stdin=inf, stdout=File(outf, "a"), stderr=None)
        pl.wait()
        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "cat <.*/input/simple1.txt \\| cat >.*/output/test_pipettor.PipelineTests.testAppendFile.out$", isRe=True)

    def __bogusStdioExpectRe(self):
        return "^invalid stdio specification object type: <class 'float'> 3\\.14159$"

    def testBogusStdin(self):
        # test stdin specification is not legal
        nopen = self.numOpenFiles()
        with self.assertRaisesRegex(PipettorException, self.__bogusStdioExpectRe()):
            pl = Pipeline([("date",), ("date",)], stdin=3.14159)
            pl.wait()
        self.orphanChecks(nopen)

    def testBogusStdout(self):
        # test stdout specification is not legal
        nopen = self.numOpenFiles()
        with self.assertRaisesRegex(PipettorException, self.__bogusStdioExpectRe()):
            pl = Pipeline([("date",), ("date",)], stdout=3.14159)
            pl.wait()
        self.orphanChecks(nopen)

    def testBogusStderr(self):
        # test stderr specification is not legal
        nopen = self.numOpenFiles()
        with self.assertRaisesRegex(PipettorException, self.__bogusStdioExpectRe()):
            pl = Pipeline([("date",), ("date",)], stderr=3.14159)
            pl.wait()
        self.orphanChecks(nopen)

    def testDataReaderBogusShare(self):
        # test stderr specification is not legal
        nopen = self.numOpenFiles()
        dr = DataReader()
        with self.assertRaisesRegex(PipettorException, "^DataReader already bound to a process$"):
            pl = Pipeline([("date",), ("date",)], stdout=dr, stderr=dr)
            pl.wait()
        self.orphanChecks(nopen)

    def testDataWriterBogusShare(self):
        # test stderr specification is not legal
        nopen = self.numOpenFiles()
        dw = DataWriter("fred")
        with self.assertRaisesRegex(PipettorException, "^DataWriter already bound to a process$"):
            pl1 = Pipeline([("cat", "/dev/null"), ("cat", "/dev/null")], stdin=dw)
            Pipeline([("cat", "/dev/null"), ("cat", "/dev/null")], stdin=dw)
        pl1.shutdown()  # clean up unstarted process
        self.orphanChecks(nopen)

    def testIntArg(self):
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        dr = DataReader()
        pl = Pipeline(("head", -2), stdin=inf, stdout=dr)
        pl.wait()
        self.assertEqual(dr.data, "one\ntwo\n")
        self.commonChecks(nopen, pl, "^head -2 <.+/input/simple1\\.txt >\\[DataReader\\]", isRe=True)


class PopenTests(PipettorTestBase):
    def __init__(self, methodName):
        super(PopenTests, self).__init__(methodName)

    def cpFileToPl(self, inName, pl):
        inf = self.getInputFile(inName)
        fh = open(inf)
        for line in fh:
            pl.write(line)
        fh.close()

    def cpPlToFile(self, pl, outExt):
        outf = self.getOutputFile(outExt)
        fh = open(outf, "w")
        for line in pl:
            fh.write(line)
        fh.close()

    def testWrite(self):
        nopen = self.numOpenFiles()
        outf = self.getOutputFile(".out")
        outfGz = self.getOutputFile(".out.gz")

        pl = Popen(("gzip", "-1"), "w", stdout=outfGz)
        self.cpFileToPl("simple1.txt", pl)
        pl.close()
        self.commonChecks(nopen, pl, "gzip -1 <.+ >.*output/test_pipettor.PopenTests.testWrite.out.gz", isRe=True)

        Pipeline(("zcat", outfGz), stdout=outf).wait()
        self.diffExpected(".out")

    def testWriteFile(self):
        nopen = self.numOpenFiles()
        outf = self.getOutputFile(".out")
        outfGz = self.getOutputFile(".out.gz")

        with open(outfGz, "w") as outfGzFh:
            pl = Popen(("gzip", "-1"), "w", stdout=outfGzFh)
            self.cpFileToPl("simple1.txt", pl)
            pl.wait()

        Pipeline(("zcat", outfGz), stdout=outf).wait()
        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "gzip -1 <.* >.*output/test_pipettor.PopenTests.testWriteFile.out.gz", isRe=True)

    def testWriteMult(self):
        nopen = self.numOpenFiles()
        outf = self.getOutputFile(".wc")

        # grr, BSD wc adds an extract space, so just convert to tabs
        pl = Popen((("gzip", "-1"),
                    ("gzip", "-dc"),
                    ("wc",),
                    ("sed", "-e", "s/  */\t/g")), "w", stdout=outf)
        self.cpFileToPl("simple1.txt", pl)
        pl.wait()

        self.diffExpected(".wc")
        self.commonChecks(nopen, pl, "^gzip -1 <.+ | gzip -dc | wc | sed -e 's/  \\*/	/g' >.*output/test_pipettor.PopenTests.testWriteMult.wc$", isRe=True)

    def testRead(self):
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        infGz = self.getOutputFile(".txt.gz")
        Pipeline(("gzip", "-c", inf), stdout=infGz).wait()

        pl = Popen(("gzip", "-dc"), "r", stdin=infGz)
        self.cpPlToFile(pl, ".out")
        pl.wait()

        self.diffExpected(".out")
        self.commonChecks(nopen, pl, "^gzip -dc <.*output/test_pipettor.PopenTests.testRead.txt.gz >.+$", isRe=True)

    def testReadMult(self):
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")

        pl = Popen((("gzip", "-1c"),
                    ("gzip", "-dc"),
                    ("wc",),
                    ("sed", "-e", "s/  */\t/g")), "r", stdin=inf)
        self.cpPlToFile(pl, ".wc")
        pl.wait()

        self.diffExpected(".wc")
        self.commonChecks(nopen, pl, "^gzip -1c <.*tests/input/simple1.txt | gzip -dc | wc | sed -e 's/  \\*/	/g' >.+$", isRe=True)

    def testExitCode(self):
        nopen = self.numOpenFiles()
        pl = Popen(("false",))
        with self.assertRaisesRegex(ProcessException, "^process exited 1: false$"):
            pl.wait()
        for p in pl.procs:
            self.assertTrue(p.returncode == 1)
        self.commonChecks(nopen, pl, "^false >.+$", isRe=True)

    simpleOneLines = ['one\n', 'two\n', 'three\n', 'four\n', 'five\n', 'six\n']

    def testReadDos(self):
        with Popen(("cat", self.getInputFile("simple1.dos.txt")), mode="r") as fh:
            lines = [l for l in fh]
        self.assertEqual(lines, self.simpleOneLines)

    def testReadMac(self):
        with Popen(("cat", self.getInputFile("simple1.mac.txt")), mode="r") as fh:
            lines = [l for l in fh]
        self.assertEqual(lines, self.simpleOneLines)

    def testSigPipe(self):
        # test not reading all of pipe output
        nopen = self.numOpenFiles()
        pl = Popen([("yes",), ("true",)], "r")
        pl.wait()
        self.commonChecks(nopen, pl, "^yes | true >.+ 2>\\[DataReader\\]$", isRe=True)

    def testReadAsAsciiReplace(self):
        # file contains unicode character outside of the ASCII range
        nopen = self.numOpenFiles()
        inf = self.getInputFile("nonAscii.txt")
        with Popen(["cat", inf], encoding='latin-1', errors="replace") as fh:
            lines = [l[:-1] for l in fh]
        self.assertEqual(['Microtubules are assembled from dimers of a- and \xc3\x9f-tubulin.'], lines)
        self.orphanChecks(nopen)


class FunctionTests(PipettorTestBase):
    def __init__(self, methodName):
        super(FunctionTests, self).__init__(methodName)

    def testWriteFile(self):
        # test write to File object
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        outf = self.getOutputFile(".out")
        run([("cat",), ("cat",)], stdin=inf, stdout=File(outf, "w"))
        self.diffExpected(".out")
        self.orphanChecks(nopen)

    def testSimplePipeFail(self):
        nopen = self.numOpenFiles()
        with self.assertRaisesRegex(ProcessException, "^process exited 1: false$"):
            run([("false",), ("true",)])
        self.orphanChecks(nopen)

    def testStdoutRead(self):
        # read from stdout into memory
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        out = runout(("sort", "-r"), stdin=inf)
        self.assertEqual(out, "two\nthree\nsix\none\nfour\nfive\n")
        self.orphanChecks(nopen)

    def testStdoutReadFail(self):
        # read from stdout into memory
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        with self.assertRaises(ProcessException) as cm:
            runout([("sort", "-r"), (os.path.join(self.getTestDir(), "progWithError"),), ("false",)], stdin=inf)
        self.checkProgWithError(cm.exception)
        self.orphanChecks(nopen)

    def testWriteFileLex(self):
        # test write to File object
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        outf = self.getOutputFile(".out")
        runlex(["cat -u", ["cat", "-u"], "cat -n"], stdin=inf, stdout=File(outf, "w"))
        self.diffExpected(".out")
        self.orphanChecks(nopen)

    def testStdoutReadLex(self):
        # read from stdout into memory
        nopen = self.numOpenFiles()
        inf = self.getInputFile("simple1.txt")
        out = runlexout("sort -r", stdin=inf)
        self.assertEqual(out, "two\nthree\nsix\none\nfour\nfive\n")
        self.orphanChecks(nopen)


def suite():
    ts = unittest.TestSuite()
    ts.addTest(unittest.makeSuite(PipelineTests))
    ts.addTest(unittest.makeSuite(PopenTests))
    ts.addTest(unittest.makeSuite(FunctionTests))
    return ts


if __name__ == '__main__':
    unittest.main()



# Copyright 2006-2015 Mark Diekhans
import os
import sys
import unittest
import difflib
import threading
import errno
import re
import glob
import io
import logging

try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except ValueError:
    MAXFD = 256


def rmTree(root):
    "remove a file hierarchy, root can be a file or a directory"
    if os.path.isdir(root):
        for dir, subdirs, files in os.walk(root, topdown=False):
            for f in files:
                os.unlink(dir + "/" + f)
            os.rmdir(dir)
    else:
        if os.path.lexists(root):
            os.unlink(root)


def ensureDir(dir):
    """Ensure that a directory exists, creating it (and parents) if needed."""
    try:
        os.makedirs(dir)
    except OSError as ex:
        if ex.errno != errno.EEXIST:
            raise ex


def ensureFileDir(fname):
    """Ensure that the directory for a file exists, creating it (and parents) if needed.
    Returns the directory path"""
    dir = os.path.dirname(fname)
    if len(dir) > 0:
        ensureDir(dir)
        return dir
    else:
        return "."


class LoggerForTests():
    """test logger that logs to memory, each instance has a new logger"""
    def __init__(self, level=logging.DEBUG):
        self.logger = logging.getLogger(str(id(self)))
        self.logger.setLevel(level)
        self.__buffer = io.StringIO()
        self.logger.addHandler(logging.StreamHandler(self.__buffer))

    @property
    def data(self):
        return self.__buffer.getvalue()


class TestCaseBase(unittest.TestCase):
    """Base class for test case with various test support functions"""

    def __init__(self, methodName):
        """initialize, removing old output files associated with the class"""
        super(TestCaseBase, self).__init__(methodName=methodName)
        clId = self.getClassId()
        od = self.getOutputDir()
        for f in glob.glob("{}/{}.*".format(od, clId)) + glob.glob("{}/tmp.*{}.*".format(od, clId)):
            rmTree(f)

    def getClassId(self):
        """Get the first part of the portable test id, consisting
        moduleBase.class.  This is the prefix to output files"""
        # module name is __main__ when run standalone, so get base file name
        mod = os.path.splitext(os.path.basename(sys.modules[self.__class__.__module__].__file__))[0]
        return mod + "." + self.__class__.__name__

    def getId(self):
        """get the fixed test id, which is in the form moduleBase.class.method
        which avoids different ids when test module is run as main or
        from a larger program"""
        # last part of unittest id is method
        return self.getClassId() + "." + self.id().split(".")[-1]

    def getTestDir(self):
        """find test directory, where concrete class is defined."""
        testDir = os.path.dirname(sys.modules[self.__class__.__module__].__file__)
        if testDir == "":
            testDir = "."
        testDir = os.path.realpath(testDir)
        # turn this into a relative directory
        cwd = os.getcwd()
        if testDir.startswith(cwd):
            testDir = testDir[len(cwd) + 1:]
            if len(testDir) == 0:
                testDir = "."
        return testDir

    def getTestRelProg(self, progName):
        "get path to a program in directory above the test directory"
        return os.path.join(self.getTestDir(), "..", progName)

    def getInputFile(self, fname):
        """Get a path to a file in the test input directory"""
        return self.getTestDir() + "/input/" + fname

    def getOutputDir(self):
        """get the path to the output directory to use for this test, create if it doesn't exist"""
        d = self.getTestDir() + "/output"
        ensureDir(d)
        return d

    def getOutputFile(self, ext):
        """Get path to the output file, using the current test id and append
        ext, which should contain a dot"""
        f = self.getOutputDir() + "/" + self.getId() + ext
        ensureFileDir(f)
        return f

    def getExpectedFile(self, ext, basename=None):
        """Get path to the expected file, using the current test id and append
        ext. If basename is used, it is inset of the test id, allowing share
        an expected file between multiple tests."""
        return self.getTestDir() + "/expected/" + (basename if basename is not None else self.getId()) + ext

    def __getLines(self, file):
        with open(file) as fh:
            return fh.readlines()

    def __getBytes(self, file):
        with open(file, "rb") as fh:
            return fh.read()

    def mustExist(self, path):
        if not os.path.exists(path):
            self.fail("file does not exist: " + path)

    def diffExpected(self, ext, expectedBasename=None):
        """diff expected and output files.  If expectedBasename is used, it is
        used insted of the test id to find the expected file, allowing share
        an expected file between multiple tests."""

        expFile = self.getExpectedFile(ext, expectedBasename)
        expLines = self.__getLines(expFile)

        outFile = self.getOutputFile(ext)
        outLines = self.__getLines(outFile)

        diff = difflib.unified_diff(expLines, outLines, expFile, outFile)
        cnt = 0
        for l in diff:
            sys.stdout.write(l)
            cnt += 1
        self.assertTrue(cnt == 0)

    def diffBinaryExpected(self, ext, expectedBasename=None):
        """diff expected and output binary files.  If expectedBasename is
        used, it is used insted of the test id to find the expected file,
        allowing share an expected file between multiple tests."""

        expFile = self.getExpectedFile(ext, expectedBasename)
        expBytes = self.__getBytes(expFile)

        outFile = self.getOutputFile(ext)
        outBytes = self.__getBytes(outFile)

        self.assertEqual(outBytes, expBytes)

    def createOutputFile(self, ext, contents=""):
        """create an output file, filling it with contents."""
        fpath = self.getOutputFile(ext)
        ensureFileDir(fpath)
        fh = open(fpath, "w")
        try:
            fh.write(contents)
        finally:
            fh.close()

    def verifyOutputFile(self, ext, expectContents=""):
        """verify an output file, if contents is not None, it is a string to
        compare to the expected contents of the file."""
        fpath = self.getOutputFile(ext)
        self.mustExist(fpath)
        self.assertTrue(os.path.isfile(fpath))
        fh = open(fpath)
        try:
            got = fh.read()
        finally:
            fh.close()
        self.assertEqual(got, expectContents)

    @staticmethod
    def numRunningThreads():
        "get the number of threads that are running"
        n = 0
        for t in threading.enumerate():
            if t.is_alive():
                n += 1
        return n

    def assertSingleThread(self):
        "fail if more than one thread is running"
        self.assertEqual(self.numRunningThreads(), 1)

    def assertNoChildProcs(self):
        "fail if there are any running or zombie child process"
        foundChild = True
        try:
            s = os.waitpid(0, os.WNOHANG)
        except OSError as ex:
            if ex.errno != errno.ECHILD:
                raise
            foundChild = False
        if foundChild:
            self.fail("pending child processes or zombies: " + str(s))

    @staticmethod
    def numOpenFiles():
        "count the number of open files"
        n = 0
        for fd in range(0, MAXFD):
            try:
                os.fstat(fd)
            except OSError:
                n += 1
        return MAXFD - n

    def assertNumOpenFilesSame(self, prevNumOpen):
        "assert that the number of open files has not changed"
        numOpen = self.numOpenFiles()
        if numOpen != prevNumOpen:
            self.fail("number of open files changed, was " + str(prevNumOpen) + ", now it's " + str(numOpen))

    def assertRegexpMatchesDotAll(self, obj, expectRe, msg=None):
        """Fail if the str(obj) does not match expectRe operator, including `.' matching newlines"""
        if not re.match(expectRe, str(obj), re.DOTALL):
            raise self.failureException(msg or "'{}' does not match '{}'".format(str(obj), expectRe))



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import setuptools


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
]

setuptools.setup(
    name = 'pipettor',
    version = '0.7.0',
    description = "pipettor - robust, easy to use Unix process pipelines",
    long_description = readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    author = "Mark Diekhans",
    author_email = 'markd@ucsc.edu',
    url = 'https://github.com/diekhans/pipettor',
    packages = [
        'pipettor',
    ],
    package_dir = {'': 'lib'},
    include_package_data = True,
    install_requires = requirements,
    license = "MIT",
    zip_safe = True,
    keywords = ['Unix', 'process', 'pipe'],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    python_requires='>=3.7',
)



# Copyright 2006-2015 Mark Diekhans
"""
pipettor interfaces to files and pipes, as well as some other IPC stuff.
"""
import os
import errno
import threading
from pipettor.exceptions import PipettorException


# note:
# A problem with python threads and signal handling is that SIGINT (C-c) will
# not raise an exception if it's blocked in thread join.  If the thread never
# terminates, the process hangs, not responding to SIGINT.  This can happen if
# the forked process is hung.  to work around this, setting the I/O threads to
# daemon solves the problem.  It also cause the process to to wait if the main
# process exists and close hasn't been called.
#
#  http://bugs.python.org/issue21822
#  http://code.activestate.com/recipes/496735-workaround-for-missed-sigint-in-multithreaded-prog/


class Dev(object):
    """Base class for objects specifying process input or output.  They
    provide a way of hide details of setting up interprocess
    communication.

    Derived class implement the following properties, if applicable:
       read_fd - file integer descriptor for reading
       read_fh - file object for reading
       write_fd - file integer descriptor for writing
       write_fh - file object for writing"""

    def _bind_read_to_process(self, process):
        """associate read side with child process."""
        pass

    def _bind_write_to_process(self, process):
        """associate write side with child process."""
        pass

    def _bind_to_process(self, process, mode):
        """associate with a child process based on mode"""
        if mode.startswith("r"):
            self._bind_read_to_process(process)
        else:
            self._bind_write_to_process(process)

    def _post_start_parent(self):
        "called do any post-exec handling in the parent"
        pass

    def close(self):
        """close the device"""
        pass


class DataReader(Dev):
    """Object to asynchronously read data from process into memory via a pipe.  A
    thread is use to prevent deadlock when both reading and writing to a child
    pipeline.

    Specifying binary access results in data of type bytes, otherwise str type
    is returned.  The buffering, encoding, and errors arguments are as used in
    the open() function.
    """
    def __init__(self, binary=False, buffering=-1, encoding=None, errors=None):
        super(DataReader, self).__init__()
        self.binary = binary
        self._process = None
        self._buffer = []
        self._thread = None
        self.read_fh = self.write_fd = None
        read_fd, self.write_fd = os.pipe()
        mode = "rb" if binary else "r"
        self.read_fh = open(read_fd, mode, buffering=buffering, encoding=encoding, errors=errors)

    def __str__(self):
        return "[DataReader]"

    def _bind_write_to_process(self, process):
        """associate write side with child process."""
        if self._process is not None:
            raise PipettorException("DataReader already bound to a process")
        self._process = process

    def _post_start_parent(self):
        "called to do any post-start handling in the parent"
        os.close(self.write_fd)
        self.write_fd = None

        self._thread = threading.Thread(target=self._reader)
        self._thread.daemon = True  # see note at top of this file
        self._thread.start()

    def close(self):
        "close pipes and terminate thread"
        if self._thread is not None:
            self._thread.join()
            self._thread = None
        if self.read_fh is not None:
            self.read_fh.close()
            self.read_fh = None
        if self.write_fd is not None:
            os.close(self.write_fd)
            self.write_fd = None

    def _reader(self):
        "child read thread function"
        assert self.write_fd is None
        self._buffer.append(self.read_fh.read())

    @property
    def data(self):
        "return buffered data as a string or bytes"
        if self.binary:
            return b"".join(self._buffer)
        else:
            return "".join(self._buffer)


class DataWriter(Dev):
    """Object to asynchronously write data to process from memory via a pipe.  A
    thread is use to prevent deadlock when both reading and writing to a child
    pipeline.  Text or binary output is determined by the type of data.

    The buffering, encoding, and errors arguments are as used in
    the open() function.
    """

    def __init__(self, data, buffering=-1, encoding=None, errors=None):
        super(DataWriter, self).__init__()
        binary = not isinstance(data, str)
        self._data = data
        self._thread = None
        self._process = None
        self.read_fd = self.write_fh = None
        self.read_fd, write_fd = os.pipe()
        mode = "wb" if binary else "w"
        self.write_fh = open(write_fd, mode, buffering=buffering, encoding=encoding, errors=errors)

    def __str__(self):
        return "[DataWriter]"

    def _bind_read_to_process(self, process):
        """associate write side with child process."""
        if self._process is not None:
            raise PipettorException("DataWriter already bound to a process")
        self._process = process

    def _post_start_parent(self):
        "called to do any start-exec handling in the parent"
        os.close(self.read_fd)
        self.read_fd = None

        self._thread = threading.Thread(target=self._writer)
        self._thread.daemon = True  # see note at top of this file
        self._thread.start()

    def close(self):
        "close pipes and terminate thread"
        if self._thread is not None:
            self._thread.join()
            self._thread = None
        if self.read_fd is not None:
            os.close(self.read_fd)
            self.read_fd = None
        if self.write_fh is not None:
            self.write_fh.close()
            self.write_fh = None

    def _writer(self):
        "write thread function"
        assert self.read_fd is None
        try:
            self.write_fh.write(self._data)
            self.write_fh.close()
            self.write_fh = None
        except IOError as ex:
            # don't raise error on broken pipe
            if ex.errno != errno.EPIPE:
                raise


class File(Dev):
    """A file path for input or output, used for specifying stdio associated
    with files. Mode is invalued on of standard r, w, or a"""

    def __init__(self, path, mode="r"):
        super(File, self).__init__()
        self.path = path
        self.mode = mode
        # only one of the file descriptors is ever opened
        self.read_fd = self.write_fd = None
        if mode.find('r') >= 0:
            self.read_fd = os.open(self.path, os.O_RDONLY)
        elif mode.find('w') >= 0:
            self.write_fd = os.open(self.path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, 0o666)
        elif mode.find('a') >= 0:
            self.write_fd = os.open(self.path, os.O_WRONLY | os.O_CREAT | os.O_APPEND, 0o666)
        else:
            raise PipettorException("invalid or unsupported mode '{}' opening {}".format(mode, path))

    def __str__(self):
        return self.path

    def close(self):
        "close file if open"
        if self.read_fd is not None:
            os.close(self.read_fd)
            self.read_fd = None
        if self.write_fd is not None:
            os.close(self.write_fd)
            self.write_fd = None

    def _post_start_parent(self):
        """post-fork child setup."""
        self.close()


class _SiblingPipe(Dev):
    """Interprocess communication between two child process by anonymous
    pipes."""

    def __init__(self):
        super(_SiblingPipe, self).__init__()
        self.read_fd, self.write_fd = os.pipe()

    def __str__(self):
        return "[Pipe]"

    def _post_start_parent(self):
        "called to do any post-exec handling in the parent"
        os.close(self.read_fd)
        self.read_fd = None
        os.close(self.write_fd)
        self.write_fd = None

    def close(self):
        if self.read_fd is not None:
            os.close(self.read_fd)
            self.read_fd = None
        if self.write_fd is not None:
            os.close(self.write_fd)
            self.write_fd = None
